import jcreek.cke.importexport.xml.*;
import jcreek.cke.importexport.KnowledgeModelImportUtilities;
import jcreek.representation.*;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * XML-parser class that does not take status windows into account.
 * Code adapted from jcreek.cke.importexport.jcxml.JCXMLImportParser
 * 
 * The one single task of this parser is to parse incoming cases extracted from
 * syntactical data of ultrasound images.
 * @author jano
 * @see jcreek.cke.importexport.jcxml.JCXMLImportParser
 */
public class CaseParser{
	private boolean abortOnFatalErrors = false;
	private boolean fatalError = false;
	private Document document = null;
	private KnowledgeModel km;
	private KnowledgeModelImportUtilities kmiu;
	
	public CaseParser(Document document, KnowledgeModel km, boolean abortOnFatalErrors){
		this.km = km;
		this.document = document;
		this.abortOnFatalErrors = abortOnFatalErrors;
		kmiu = new KnowledgeModelImportUtilities(km);
	}
	
	/**
	 * Starts the traversal of the DOM document.
	 */
	public void startTraverse() {
		try {
			traverse(document);
		} catch (Exception ex) {
			fatalError = true;
            ex.printStackTrace();
			System.out.println("An exception was caught dring the traversal of the DOM document!");
		}
	} // startTraverse
	
	/**
	 * Checks if abortOnFatalErrors is enabled or not.
	 *
	 * @return true if bortOnFatalErrors is enabled, otherwise false.
	 */
	public boolean isAbortOnFatalErrorsEnabled() {
		return abortOnFatalErrors;
	} // isAbortOnFatalErrorsEnabled

	/**
	 * Checks if an fatal error did occur.
	 *
	 * @return true if an fatal error did occur, otherwise false.
	 */
	public boolean didFatalErrorOccur() {
		return fatalError;
	} // didFatalErrorOccur

	/**
	 * A recursive method that traverses the Document and locates the
	 * different Elements needed to collect the knowledge model
	 * contained in the XML file.
	 *
	 * @param node the node with childs to traverse. If the node is null
	 * no traversal will be done.
	 */
	private void traverse(Node node) throws Exception {
		// is there anything to do?
		if ((node == null) && (fatalError && abortOnFatalErrors)) {
			return;
		}
        //Vector entityList = new Vector();
        //Vector relationList = new Vector();
        //Vector caseList = new Vector();

		int type = node.getNodeType();
		switch (type) {
			// handle Document node, that is only to traverse the children...
			case Node.DOCUMENT_NODE :
			{
				traverse(((Document) node).getDocumentElement());
				break;
			} // case Node.DOCUMENT_NODE :
			// handle the Element node, this is where the main part is
			case Node.ELEMENT_NODE :
			{
				// handle the About element
				if (node.getNodeName().equalsIgnoreCase("about")) {
					handleAboutElement(node);
					// insert a return here so that the child elements of
					// the About element isn't handled more than once.
					return;
				} else if (node.getNodeName().equalsIgnoreCase("newRelationType")) {
					handleNewRelationElement(node);
					return;
				} else if (node.getNodeName().equalsIgnoreCase("entity")) {
					if (!fatalError) {
						//entityList.add(node);
						handleEntityElement(node);
					}
					return;
				} else if (node.getNodeName().equalsIgnoreCase("relation")) {
					if (!fatalError) {
						//relationList.add(node);
						handleRelationElement(node);
					}
					return;
				} else if (node.getNodeName().equalsIgnoreCase("case")) {
					if (!fatalError) {
						//caseList.add(node);
						handleCaseElement(node);
					}
					// insert a return here so that the child elements of
					// the Case element isn't handled more than once.
					return;
				} else {
					// Elements here might eg. be 'jcreekmodel', 'model', ...
					Node child = node.getFirstChild();
					while ((child != null) && !(fatalError && abortOnFatalErrors)) {
						traverse(child);
						child = child.getNextSibling();
					}
				}
			} // case Node.ELEMENT_NODE
		} // switch(type)
	} // traverse

	/**
	 * Does the handling of the general information about the knowledge model.
	 * <p>
	 * The About element contains all the basic information of the
	 * Knowledge Model in this JCXML file.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleAboutElement(Node node) {
		/**************************************
		<about>
			<title>Car Example</title>
			<name>CarExample</name>
			<creator>Stein L. Tomassen</creator>
			<date>2002-05-23</date>
			<description>A car example knowledge model</description>
			<language>en</language>
		</about>
		***************************************/
		Node child = node.getFirstChild();
		while (child != null) {
			if (child.getNodeType() == Node.ELEMENT_NODE) {
				Node grandChild = child.getFirstChild();
				while (grandChild != null) {
					if (grandChild.getNodeType() == Node.TEXT_NODE) {
						if (child.getNodeName().equalsIgnoreCase("title")) {
                            System.out.println("km: "+km+ "gc: "+grandChild);
							km.setTitle(grandChild.getNodeValue());
						} else if (child.getNodeName().equalsIgnoreCase("name")) {
							km.setName(grandChild.getNodeValue());
						} else if (child.getNodeName().equalsIgnoreCase("creator")) {
						} else if (child.getNodeName().equalsIgnoreCase("date")) {
						} else if (child.getNodeName().equalsIgnoreCase("description")) {
							km.setDescription(grandChild.getNodeValue());
						} else if (child.getNodeName().equalsIgnoreCase("language")) {
						}
					}
					grandChild = grandChild.getNextSibling();
				}
			} // if (child.getNodeType()...
			child = child.getNextSibling();
		}
	} // handleAboutElement

	/**
	 * Does the handling of new relations.
	 * <p>
	 * If a 'classDescriptor' does not exists, it then displays a fatal error message
	 * in the process output view. What it could have done is to put the 'new relation'
	 * aside for handling later on, but since all new relations are a subclass of the
	 * 'basic relations' and the new relations will not add any new functionality then
	 * I don't quite see the point. Instead there is a demand that all new relations
	 * should be a direct subclass of the 'basic relations.
	 * <p>
	 * If a relation already exists then the method will display a warning that the
	 * relation already exists.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleNewRelationElement(Node node) {
		/**************************************
		<newRelationType classDescriptor="has finding" name="has colour"
		 inverseName="colour of" defaultStrength="0.5"
		 description="describes the color of an entity"/>
		***************************************/
		// 'defaultStrength' and 'description' are not mandatory

		String classDescriptor = DOMutilities.getAttributeValueByName("classDescriptor", node.getAttributes());
		String name = DOMutilities.getAttributeValueByName("name", node.getAttributes());
		String inverseName = DOMutilities.getAttributeValueByName("inverseName", node.getAttributes());
		String defaultStrength = DOMutilities.getAttributeValueByName("defaultStrength", node.getAttributes());
		String description = DOMutilities.getAttributeValueByName("description", node.getAttributes());

		// now add the new relation
		try {
			kmiu.addNewRelationType(classDescriptor, name, inverseName, defaultStrength, description);
		} catch (KnowledgeModelImportUtilities.CompilationFault ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
		} catch (KnowledgeModelImportUtilities.EqualRelationTypeAlreadyExist ex) {
			//ex.printStackTrace();
			//System.out.println("Exception caught! An equal relation type does already exist!");
		} catch (KnowledgeModelImportUtilities.ClassDescriptorDoNotExist ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
		} catch (KnowledgeModelImportUtilities.RelationTypeAlreadyExist ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}

		} catch (KnowledgeModelImportUtilities.RelationTypeNameAlreadyInUse ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
		}
	} // handleNewRelationElement

	/**
	 * Does the handling of the entities.
	 * <p>
	 * If an entity already exists a warning message will be displayed.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleEntityElement(Node node) {
		/**************************************
		<entity name="engine status" description="engine status"/>
		***************************************/
		// 'description' is not mandatory
		String name = DOMutilities.getAttributeValueByName("name", node.getAttributes());
		String description = DOMutilities.getAttributeValueByName("description", node.getAttributes());
        String valueClass = DOMutilities.getAttributeValueByName("valueClass", node.getAttributes());
        String value = DOMutilities.getAttributeValueByName("value", node.getAttributes());

		// now add the new entity
		try {
			kmiu.addEntity(name, description, valueClass, value);
		} catch (KnowledgeModelImportUtilities.EntityAlreadyExist ex) {
			//ex.printStackTrace();
			//System.out.println("Exception caught! The entity does already exist!");
		}
	} // handleEntityElement

	/**
	 * Does the handling of the relations among the entities.
	 * <p>
	 * If a relation already exist a warning message will be displayed. If
	 * the relation refers to an entity or a relation type that does not
	 * exist a fatal error message will be displayed.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleRelationElement(Node node) {
		/**************************************
		<relation fromEntity="battery status" relationType="has instance" toEntity="OK battery" strength="0.5"/>
		***************************************/
		// 'strength' is not mandatory
		String fromEntity = DOMutilities.getAttributeValueByName("fromEntity", node.getAttributes());
		String relationType = DOMutilities.getAttributeValueByName("relationType", node.getAttributes());
		String toEntity = DOMutilities.getAttributeValueByName("toEntity", node.getAttributes());
		String strength = DOMutilities.getAttributeValueByName("strength", node.getAttributes());

		try {
			kmiu.addRelation(fromEntity, relationType, toEntity, strength);
		} catch (KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex) {
			//ex.printStackTrace();
			//System.out.println("Exception caught! The entity relation does already exist!");
		} catch (KnowledgeModelImportUtilities.EntityDoNotExist ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
		} catch (KnowledgeModelImportUtilities.RelationTypeDoNotExist ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
		}
	} // handleNewRelationElement

	/**
	 * Does the handling of the cases.
	 * <p>
	 * Does in the same process take care of the findings and the possible solution
	 * of the case as well. If the case doesn't have a solution the status of the
	 * case will be set to 'proccessed'.
	 * <p>
	 * Displayes a warning message if the case already exist. Does also display a
	 * warning message if a finding or solution already exist. If the finding or
	 * solution refers to an entity or a relation type that does not exist a fatal
	 * error message will be displayed.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleCaseElement(Node node) {
		//System.out.println("Found new case!");
		//		statusWindow.appendText("Handling Case element...");
		/**************************************
		<?xml version="1.0" encoding="ISO-8859-1"?>
		<root xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="jcreekmodel.xsd">
			<case description="test" name="testCase1">
			  <finding relationType="belongs to" toEntity="mad"/>
			  <finding relationType="has landmark" toEntity="ribs_left">
			  <finding relationType="has area" toEntity="345"/>
			  <finding relationType="has compactness" toEntity="24"/>
			  </region>
			</case>
		</root>
		***************************************/
		// 'finding', 'importance', 'predictiveStrength', 'solution', 'solution:toEntity',
		// 'solution:predictiveStrength', and 'solution:importance' are not mandatory
		// 'status' might either be 'solved' or 'unsolved', if no solution element then the case
		// is in 'processing' state

		// also handle the subelements
		// but first handle the attributes
		String name = DOMutilities.getAttributeValueByName("name", node.getAttributes());
		String description = DOMutilities.getAttributeValueByName("description", node.getAttributes());
		
		name = "objectCase#"+(int)(100*Math.random());

		while (km.getEntity(name) != null) {
			name = "objectCase#"+(int)(100*Math.random());
		}
		
			// now create the case
		try {
			kmiu.addCase(name, description);
			System.out.println("New case '"+ name +"' added to the knowledge model!");
		} catch (KnowledgeModelImportUtilities.CaseAlreadyExist ex) {
			//ex.printStackTrace();
			System.out.println("Exception caught! The case "+ name +" does already exist!");
		}

		// even if the case should already exists there is no problem to continue
		// it might be that new findings or solution could be added.
		// now continue with the findings and a possible solution
		Node child = node.getFirstChild();
		while (child != null && !(fatalError && abortOnFatalErrors)) {
			if (child.getNodeType() == Node.ELEMENT_NODE) {
				if (child.getNodeName().equalsIgnoreCase("finding")) {
					// first collect the attrbutes
					String relationType = DOMutilities.getAttributeValueByName("relationType", child.getAttributes());
					String toEntity = DOMutilities.getAttributeValueByName("toEntity", child.getAttributes());
					String importance = DOMutilities.getAttributeValueByName("importance", child.getAttributes());
					String predictiveStrength = DOMutilities.getAttributeValueByName("predictiveStrength",child.getAttributes());
					
					// If the parsed finding has a numerical value, create a new NumberEntity
					//if((relationType.length()>11 && (relationType.substring(relationType.length()-11)).equalsIgnoreCase("measurement"))){
					//	toEntity = createNumberEntity(relationType,toEntity);
					//}
						// then add the finding
					try {
						kmiu.addHasFinding(name, relationType, toEntity, importance, predictiveStrength);
						System.out.println("Added new finding '"+ relationType +"' to case "+name);
					} catch (KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex) {
						System.out.println("Exception caught, an equal relation to "+relationType+" does already exist.");
					} catch (KnowledgeModelImportUtilities.EntityDoNotExist ex) {
						if (abortOnFatalErrors) {
							fatalError = true;
						}
						System.out.println("Exception caught, entity "+ toEntity +" does not exist.");
					} catch (KnowledgeModelImportUtilities.RelationTypeDoNotExist ex) {
						if (abortOnFatalErrors) {
							fatalError = true;
						}
						System.out.println("Exception caught, relation type "+ relationType +" do not exist.");
					}
				}
				/*
				// landmarks in the viewplane to be added to the case
				else if (child.getNodeName().equalsIgnoreCase("region")) {
					String relationType = "has landmark"; // We already know that this is a landmark of the viewplane
					String regionName = DOMutilities.getAttributeValueByName("name", child.getAttributes());
					String importance = DOMutilities.getAttributeValueByName("importance",child.getAttributes());
					String predictiveStrength = DOMutilities.getAttributeValueByName("predictiveStrength",child.getAttributes());
					
					try{
						kmiu.addHasFinding(name, relationType, regionName, importance, predictiveStrength);
						System.out.println("New landmark "+ regionName +" successfully added to the case!");
					} catch (KnowledgeModelImportUtilities.EntityDoNotExist ex) {
						if (abortOnFatalErrors) {
							fatalError = true;
						}
						System.out.println("Exception caught, landmark "+ regionName +" does not exist.");
					} catch (KnowledgeModelImportUtilities.RelationTypeDoNotExist ex) {
						if (abortOnFatalErrors) {
							fatalError = true;
						}
						System.out.println("Exception caught, relation type "+ relationType +" do not exist.");
					} catch (KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex) {
						System.out.println("Exception caught, an equal relation to "+relationType+" does already exist.");
					}
					// findings related to the specified landmark
					Node grandChild = child.getFirstChild();
					while (grandChild != null && !(fatalError && abortOnFatalErrors)) {
						if (grandChild.getNodeType() == Node.ELEMENT_NODE) {
							if (grandChild.getNodeName().equalsIgnoreCase("finding")) {
								String relationTypeR = DOMutilities.getAttributeValueByName("relationType", grandChild.getAttributes());
								String toEntityR = DOMutilities.getAttributeValueByName("toEntity", grandChild.getAttributes());
								String importanceR = DOMutilities.getAttributeValueByName("importance", grandChild.getAttributes());
								String predictiveStrengthR = DOMutilities.getAttributeValueByName("predictiveStrength",grandChild.getAttributes());
							
								String entityName = createNumberEntity(relationTypeR,toEntityR);
								try{
									kmiu.addHasFinding(regionName, relationTypeR, entityName, importanceR, predictiveStrengthR);
								} catch (KnowledgeModelImportUtilities.EntityDoNotExist ex) {
									if (abortOnFatalErrors) {
										fatalError = true;
									}
									System.out.println("Exception caught, entity "+ toEntityR +" does not exist.");
								} catch (KnowledgeModelImportUtilities.RelationTypeDoNotExist ex) {
									if (abortOnFatalErrors) {
										fatalError = true;
									}
									System.out.println("Exception caught, relation type "+ relationType +" do not exist.");
								} catch (KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex) {
									System.out.println("Exception caught, an equal relation to "+relationTypeR+" does already exist.");
								}
							}
						}
						grandChild = grandChild.getNextSibling(); // Get the next finding of the region
					}
				}*/
				// the possible solution
				else if (child.getNodeName().equalsIgnoreCase("solution")) {
					// first collect the attributes
					String status = DOMutilities.getAttributeValueByName("status", child.getAttributes());
					String toEntity = DOMutilities.getAttributeValueByName("toEntity",child.getAttributes());
					String importance =DOMutilities.getAttributeValueByName("importance",child.getAttributes());
					String predictiveStrength =DOMutilities.getAttributeValueByName("predictiveStrength", child.getAttributes());

					// then add the solution
					try {
						kmiu.addHasSolution(name,status, toEntity, importance, predictiveStrength);
						System.out.println("Added new solution '"+ toEntity +"' to case "+name);
					} catch (KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex) {
						System.out.println("Exception caught, an equal relation to "+status+" does already exist.");
					} catch (KnowledgeModelImportUtilities.EntityDoNotExist ex) {
						if (abortOnFatalErrors) {
							fatalError = true;
						}
						System.out.println("Exception caught, entity "+ toEntity +" does not exist.");
					}
				}
			} // if (child.getNodeType()...
			child = child.getNextSibling();
		} // while
	} // handleCaseElement
	
	/**
	 * Create a new NumberEntity on the fly and add it to the knowledge model
	 * 
	 * @param entity 	NumberEntity created as a value of this entity
	 * @param value		Value of the new NumberEntity
	 * @return			The name of the new entity
	 */
	public String createNumberEntity(String entity, String value){
		// Use the relationType to decide what kind of entity we are dealing with
		// Example: "has area" --> area
		Entity newEntity = km.getEntity(entity.substring(4));
		NumberEntity newNumberEntity = new NumberEntity(km, new Double(Double.parseDouble(value)), newEntity);
		return newNumberEntity.getName();
	}
} // class JCXMLimportParser